# Copyright (c) 2021, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# Licensed under the MIT License [see LICENSE for details].

import numpy as np

_DTYPE = np.dtype(
    {
        "names": [
            "body_id_a",
            "body_id_b",
            "link_id_a",
            "link_id_b",
            "position_a_world",
            "position_b_world",
            "position_a_link",
            "position_b_link",
            "normal",
            "force",
        ],
        "formats": [
            np.int32,
            np.int32,
            np.int32,
            np.int32,
            [("x", np.float32), ("y", np.float32), ("z", np.float32)],
            [("x", np.float32), ("y", np.float32), ("z", np.float32)],
            [("x", np.float32), ("y", np.float32), ("z", np.float32)],
            [("x", np.float32), ("y", np.float32), ("z", np.float32)],
            [("x", np.float32), ("y", np.float32), ("z", np.float32)],
            np.float32,
        ],
        "offsets": [0, 4, 8, 12, 16, 28, 40, 52, 64, 76],
        "itemsize": 80,
    }
)


def create_contact_array(
    length,
    body_id_a=None,
    body_id_b=None,
    link_id_a=None,
    link_id_b=None,
    position_a_world=None,
    position_b_world=None,
    position_a_link=None,
    position_b_link=None,
    normal=None,
    force=None,
):
    """ """
    array = np.empty(length, dtype=_DTYPE)
    if length > 0:
        array["body_id_a"] = body_id_a
        array["body_id_b"] = body_id_b
        array["link_id_a"] = link_id_a
        array["link_id_b"] = link_id_b
        array["position_a_world"] = position_a_world
        array["position_b_world"] = position_b_world
        array["position_a_link"] = position_a_link
        array["position_b_link"] = position_b_link
        array["normal"] = normal
        array["force"] = force
    return array



# Copyright (c) 2021, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# Licensed under the MIT License [see LICENSE for details].

import abc


class Simulator(abc.ABC):
    """Simulator."""

    def __init__(self, cfg, scene):
        """ """
        self._cfg = cfg
        self._scene = scene

    @abc.abstractmethod
    def reset(self, env_ids):
        """ """

    @abc.abstractmethod
    def step(self):
        """ """

    @property
    @abc.abstractmethod
    def contact(self):
        """ """

    @abc.abstractmethod
    def close(self):
        """ """



# Copyright (c) 2021, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# Licensed under the MIT License [see LICENSE for details].


class GeometryType:
    """ """

    URDF = 0
    SPHERE = 1
    BOX = 2


class DoFControlMode:
    """ """

    NONE = 0
    POSITION_CONTROL = 1
    VELOCITY_CONTROL = 2
    TORQUE_CONTROL = 3


class MeshNormalMode:
    """ """

    FROM_ASSET = 0
    COMPUTE_PER_VERTEX = 1
    COMPUTE_PER_FACE = 2



# Copyright (c) 2021, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# Licensed under the MIT License [see LICENSE for details].

import torch
import numpy as np

from contextlib import contextmanager


class Body:
    """ """

    _ATTR_TENSOR_NDIM = {
        "initial_base_position": 1,
        "initial_base_velocity": 1,
        "initial_dof_position": 1,
        "initial_dof_velocity": 1,
        "dof_target_position": 1,
        "dof_target_velocity": 1,
        "dof_force": 1,
    }
    _ATTR_ARRAY_NDIM = {
        "scale": 0,
        "link_collision_filter": 1,
        "link_lateral_friction": 1,
        "link_spinning_friction": 1,
        "link_rolling_friction": 1,
        "link_restitution": 1,
        "link_linear_damping": 0,
        "link_angular_damping": 0,
        "link_color": 2,
        "link_segmentation_id": 1,
        "dof_has_limits": 1,
        "dof_lower_limit": 1,
        "dof_upper_limit": 1,
        "dof_control_mode": 0,
        "dof_max_force": 1,
        "dof_max_velocity": 1,
        "dof_position_gain": 1,
        "dof_velocity_gain": 1,
        "dof_armature": 1,
    }

    _created = False

    def __init__(
        self,
        name=None,
        geometry_type=None,
        urdf_file=None,
        sphere_radius=None,
        box_half_extent=None,
        device=None,
        use_fixed_base=None,
        use_self_collision=None,
        vhacd_enabled=None,
        vhacd_params=None,
        mesh_normal_mode=None,
        env_ids_load=None,
        initial_base_position=None,
        initial_base_velocity=None,
        initial_dof_position=None,
        initial_dof_velocity=None,
        scale=None,
        link_collision_filter=None,
        link_lateral_friction=None,
        link_spinning_friction=None,
        link_rolling_friction=None,
        link_restitution=None,
        link_linear_damping=None,
        link_angular_damping=None,
        link_color=None,
        link_segmentation_id=None,
        dof_has_limits=None,
        dof_lower_limit=None,
        dof_upper_limit=None,
        dof_control_mode=None,
        dof_max_force=None,
        dof_max_velocity=None,
        dof_position_gain=None,
        dof_velocity_gain=None,
        dof_armature=None,
        dof_target_position=None,
        dof_target_velocity=None,
        dof_force=None,
    ):
        """ """
        self._init_attr_array_pipeline()
        self._init_callback()

        self.name = name
        self.geometry_type = geometry_type
        self.urdf_file = urdf_file
        self.sphere_radius = sphere_radius
        self.box_half_extent = box_half_extent
        self.device = device
        self.use_fixed_base = use_fixed_base
        self.use_self_collision = use_self_collision
        self.vhacd_enabled = vhacd_enabled
        self.vhacd_params = vhacd_params
        self.mesh_normal_mode = mesh_normal_mode
        self.env_ids_load = env_ids_load
        self.initial_base_position = initial_base_position
        self.initial_base_velocity = initial_base_velocity
        self.initial_dof_position = initial_dof_position
        self.initial_dof_velocity = initial_dof_velocity

        self.scale = scale

        self.link_collision_filter = link_collision_filter
        self.link_lateral_friction = link_lateral_friction
        self.link_spinning_friction = link_spinning_friction
        self.link_rolling_friction = link_rolling_friction
        self.link_restitution = link_restitution
        self.link_linear_damping = link_linear_damping
        self.link_angular_damping = link_angular_damping

        self.link_color = link_color
        self.link_segmentation_id = link_segmentation_id

        self.dof_has_limits = dof_has_limits
        self.dof_lower_limit = dof_lower_limit
        self.dof_upper_limit = dof_upper_limit
        self.dof_control_mode = dof_control_mode
        self.dof_max_force = dof_max_force
        self.dof_max_velocity = dof_max_velocity
        self.dof_position_gain = dof_position_gain
        self.dof_velocity_gain = dof_velocity_gain
        self.dof_armature = dof_armature

        self.dof_target_position = dof_target_position
        self.dof_target_velocity = dof_target_velocity
        self.dof_force = dof_force

        self.env_ids_reset_base_state = None
        self.env_ids_reset_dof_state = None

        self.dof_state = None
        self.link_state = None
        self.contact_id = None

        self._created = True

    def __setattr__(self, key, value):
        """ """
        # Exclude `dof_state` and `link_state` to prevent infinite recursion in property calls.
        if self._created and key not in ("dof_state", "link_state") and not hasattr(self, key):
            raise TypeError(f"Unrecognized Body attribute '{key}': {self.name}")
        object.__setattr__(self, key, value)

    def _init_attr_array_pipeline(self):
        """ """
        self._attr_array_locked = {}
        self._attr_array_dirty_flag = {}
        self._attr_array_dirty_mask = {}
        self._attr_array_default_flag = {}
        for attr in self._ATTR_ARRAY_NDIM:
            self._attr_array_locked[attr] = False
            self._attr_array_dirty_flag[attr] = False
            self._attr_array_default_flag[attr] = False

    @property
    def attr_array_locked(self):
        """ """
        return self._attr_array_locked

    @property
    def attr_array_dirty_flag(self):
        """ """
        return self._attr_array_dirty_flag

    @property
    def attr_array_dirty_mask(self):
        """ """
        return self._attr_array_dirty_mask

    @property
    def attr_array_default_flag(self):
        """ """
        return self._attr_array_default_flag

    def _init_callback(self):
        """ """
        self._callback_collect_dof_state = None
        self._callback_collect_link_state = None

    @property
    def name(self):
        """ """
        return self._name

    @name.setter
    def name(self, value):
        """ """
        self._name = value

    @property
    def geometry_type(self):
        """ """
        return self._geometry_type

    @geometry_type.setter
    def geometry_type(self, value):
        """ """
        self._geometry_type = value

    @property
    def urdf_file(self):
        """ """
        return self._urdf_file

    @urdf_file.setter
    def urdf_file(self, value):
        """ """
        self._urdf_file = value

    @property
    def sphere_radius(self):
        """ """
        return self._sphere_radius

    @sphere_radius.setter
    def sphere_radius(self, value):
        """ """
        self._sphere_radius = value

    @property
    def box_half_extent(self):
        """ """
        return self._box_half_extent

    @box_half_extent.setter
    def box_half_extent(self, value):
        """ """
        self._box_half_extent = value

    @property
    def device(self):
        """ """
        return self._device

    @device.setter
    def device(self, value):
        """ """
        self._device = value

        if hasattr(self, "_env_ids_load") and self.env_ids_load is not None:
            self.env_ids_load = self.env_ids_load.to(value)

        if hasattr(self, "_initial_base_position") and self.initial_base_position is not None:
            self.initial_base_position = self.initial_base_position.to(value)
        if hasattr(self, "_initial_base_velocity") and self.initial_base_velocity is not None:
            self.initial_base_velocity = self.initial_base_velocity.to(value)
        if hasattr(self, "_initial_dof_position") and self.initial_dof_position is not None:
            self.initial_dof_position = self.initial_dof_position.to(value)
        if hasattr(self, "_initial_dof_velocity") and self.initial_dof_velocity is not None:
            self.initial_dof_velocity = self.initial_dof_velocity.to(value)

        if hasattr(self, "_dof_target_position") and self.dof_target_position is not None:
            self.dof_target_position = self.dof_target_position.to(value)
        if hasattr(self, "_dof_target_velocity") and self.dof_target_velocity is not None:
            self.dof_target_velocity = self.dof_target_velocity.to(value)
        if hasattr(self, "_dof_force") and self.dof_force is not None:
            self.dof_force = self.dof_force.to(value)

        if hasattr(self, "_env_ids_reset_dof_state") and self.env_ids_reset_dof_state is not None:
            self.env_ids_reset_dof_state = self.env_ids_reset_dof_state.to(value)

        if hasattr(self, "_dof_state") and self.dof_state is not None:
            self.dof_state = self.dof_state.to(value)
        if hasattr(self, "_link_state") and self.link_state is not None:
            self.link_state = self.link_state.to(value)

    @property
    def use_fixed_base(self):
        """ """
        return self._use_fixed_base

    @use_fixed_base.setter
    def use_fixed_base(self, value):
        """ """
        self._use_fixed_base = value

    @property
    def use_self_collision(self):
        """ """
        return self._use_self_collision

    @use_self_collision.setter
    def use_self_collision(self, value):
        """ """
        self._use_self_collision = value

    @property
    def vhacd_enabled(self):
        """ """
        return self._vhacd_enabled

    @vhacd_enabled.setter
    def vhacd_enabled(self, value):
        """ """
        self._vhacd_enabled = value

    @property
    def vhacd_params(self):
        """ """
        return self._vhacd_params

    @vhacd_params.setter
    def vhacd_params(self, value):
        """ """
        self._vhacd_params = value

    @property
    def mesh_normal_mode(self):
        """ """
        return self._mesh_normal_mode

    @mesh_normal_mode.setter
    def mesh_normal_mode(self, value):
        """ """
        self._mesh_normal_mode = value

    @property
    def env_ids_load(self):
        """ """
        return self._env_ids_load

    @env_ids_load.setter
    def env_ids_load(self, value):
        """ """
        if value is not None:
            value = torch.as_tensor(value, dtype=torch.int64, device=self.device)
            if value.ndim != 1:
                raise ValueError("'env_ids_load' must have a number of dimensions of 1")
        self._env_ids_load = value

    @property
    def initial_base_position(self):
        """ """
        return self._initial_base_position

    @initial_base_position.setter
    def initial_base_position(self, value):
        """ """
        if value is not None:
            value = torch.as_tensor(value, dtype=torch.float32, device=self.device)
            if value.ndim not in (
                self._ATTR_TENSOR_NDIM["initial_base_position"],
                self._ATTR_TENSOR_NDIM["initial_base_position"] + 1,
            ):
                raise ValueError(
                    "'initial_base_position' must have a number of dimensions of "
                    f"{self._ATTR_TENSOR_NDIM['initial_base_position']} or "
                    f"{self._ATTR_TENSOR_NDIM['initial_base_position'] + 1}"
                )
            if value.shape[-1] != 7:
                raise ValueError("'initial_base_position' must have the last dimension of size 7")
        self._initial_base_position = value

    @property
    def initial_base_velocity(self):
        """ """
        return self._initial_base_velocity

    @initial_base_velocity.setter
    def initial_base_velocity(self, value):
        """ """
        if value is not None:
            value = torch.as_tensor(value, dtype=torch.float32, device=self.device)
            if value.ndim not in (
                self._ATTR_TENSOR_NDIM["initial_base_velocity"],
                self._ATTR_TENSOR_NDIM["initial_base_velocity"] + 1,
            ):
                raise ValueError(
                    "'initial_base_velocity' must have a number of dimensions of "
                    f"{self._ATTR_TENSOR_NDIM['initial_base_velocity']} or "
                    f"{self._ATTR_TENSOR_NDIM['initial_base_velocity'] + 1}"
                )
            if value.shape[-1] != 6:
                raise ValueError("'initial_base_velocity' must have the last dimension of size 6")
        self._initial_base_velocity = value

    @property
    def initial_dof_position(self):
        """ """
        return self._initial_dof_position

    @initial_dof_position.setter
    def initial_dof_position(self, value):
        """ """
        if value is not None:
            value = torch.as_tensor(value, dtype=torch.float32, device=self.device)
            if value.ndim not in (
                self._ATTR_TENSOR_NDIM["initial_dof_position"],
                self._ATTR_TENSOR_NDIM["initial_dof_position"] + 1,
            ):
                raise ValueError(
                    "'initial_dof_position' must have a number of dimensions of "
                    f"{self._ATTR_TENSOR_NDIM['initial_dof_position']} or "
                    f"{self._ATTR_TENSOR_NDIM['initial_dof_position'] + 1}"
                )
        self._initial_dof_position = value

    @property
    def initial_dof_velocity(self):
        """ """
        return self._initial_dof_velocity

    @initial_dof_velocity.setter
    def initial_dof_velocity(self, value):
        """ """
        if value is not None:
            value = torch.as_tensor(value, dtype=torch.float32, device=self.device)
            if value.ndim not in (
                self._ATTR_TENSOR_NDIM["initial_dof_velocity"],
                self._ATTR_TENSOR_NDIM["initial_dof_velocity"] + 1,
            ):
                raise ValueError(
                    "'initial_dof_velocity' must have a number of dimensions of "
                    f"{self._ATTR_TENSOR_NDIM['initial_dof_velocity']} or "
                    f"{self._ATTR_TENSOR_NDIM['initial_dof_velocity'] + 1}"
                )
        self._initial_dof_velocity = value

    @property
    def scale(self):
        """ """
        return self._scale

    @scale.setter
    def scale(self, value):
        """ """
        assert not self._attr_array_locked[
            "scale"
        ], "'scale' cannot be directly changed after simulation starts. Use 'update_attr_array()'."
        if value is not None:
            value = np.asanyarray(value, dtype=np.float32)
            if value.ndim not in (
                self._ATTR_ARRAY_NDIM["scale"],
                self._ATTR_ARRAY_NDIM["scale"] + 1,
            ):
                raise ValueError(
                    f"'scale' must have a number of dimensions of {self._ATTR_ARRAY_NDIM['scale']} "
                    f"or {self._ATTR_ARRAY_NDIM['scale'] + 1}"
                )
        self._scale = value

    @property
    def link_collision_filter(self):
        """ """
        return self._link_collision_filter

    @link_collision_filter.setter
    def link_collision_filter(self, value):
        """ """
        assert not self._attr_array_locked["link_collision_filter"], (
            "'link_collision_filter' cannot be directly changed after simulation starts. Use "
            "'update_attr_array()'."
        )
        if value is not None:
            value = np.asanyarray(value, dtype=np.int64)
            if value.ndim not in (
                self._ATTR_ARRAY_NDIM["link_collision_filter"],
                self._ATTR_ARRAY_NDIM["link_collision_filter"] + 1,
            ):
                raise ValueError(
                    "'link_collision_filter' must have a number of dimensions of "
                    f"{self._ATTR_ARRAY_NDIM['link_collision_filter']} or "
                    f"{self._ATTR_ARRAY_NDIM['link_collision_filter'] + 1}"
                )
        self._link_collision_filter = value

    @property
    def link_lateral_friction(self):
        """ """
        return self._link_lateral_friction

    @link_lateral_friction.setter
    def link_lateral_friction(self, value):
        """ """
        assert not self._attr_array_locked["link_lateral_friction"], (
            "'link_lateral_friction' cannot be directly changed after simulation starts. Use "
            "'update_attr_array()'."
        )
        if value is not None:
            value = np.asanyarray(value, dtype=np.float32)
            if value.ndim not in (
                self._ATTR_ARRAY_NDIM["link_lateral_friction"],
                self._ATTR_ARRAY_NDIM["link_lateral_friction"] + 1,
            ):
                raise ValueError(
                    "'link_lateral_friction' must have a number of dimensions of "
                    f"{self._ATTR_ARRAY_NDIM['link_lateral_friction']} or "
                    f"{self._ATTR_ARRAY_NDIM['link_lateral_friction'] + 1}"
                )
        self._link_lateral_friction = value

    @property
    def link_spinning_friction(self):
        """ """
        return self._link_spinning_friction

    @link_spinning_friction.setter
    def link_spinning_friction(self, value):
        """ """
        assert not self._attr_array_locked["link_spinning_friction"], (
            "'link_spinning_friction' cannot be directly changed after simulation starts. Use "
            "'update_attr_array()'."
        )
        if value is not None:
            value = np.asanyarray(value, dtype=np.float32)
            if value.ndim not in (
                self._ATTR_ARRAY_NDIM["link_spinning_friction"],
                self._ATTR_ARRAY_NDIM["link_spinning_friction"] + 1,
            ):
                raise ValueError(
                    "'link_spinning_friction' must have a number of dimensions of "
                    f"{self._ATTR_ARRAY_NDIM['link_spinning_friction']} or "
                    f"{self._ATTR_ARRAY_NDIM['link_spinning_friction'] + 1}"
                )
        self._link_spinning_friction = value

    @property
    def link_rolling_friction(self):
        """ """
        return self._link_rolling_friction

    @link_rolling_friction.setter
    def link_rolling_friction(self, value):
        """ """
        assert not self._attr_array_locked["link_rolling_friction"], (
            "'link_rolling_friction' cannot be directly changed after simulation starts. Use "
            "'update_attr_array()'."
        )
        if value is not None:
            value = np.asanyarray(value, dtype=np.float32)
            if value.ndim not in (
                self._ATTR_ARRAY_NDIM["link_rolling_friction"],
                self._ATTR_ARRAY_NDIM["link_rolling_friction"] + 1,
            ):
                raise ValueError(
                    "'link_rolling_friction' must have a number of dimensions of "
                    f"{self._ATTR_ARRAY_NDIM['link_rolling_friction']} or "
                    f"{self._ATTR_ARRAY_NDIM['link_rolling_friction'] + 1}"
                )
        self._link_rolling_friction = value

    @property
    def link_restitution(self):
        """ """
        return self._link_restitution

    @link_restitution.setter
    def link_restitution(self, value):
        """ """
        assert not self._attr_array_locked["link_restitution"], (
            "'link_restitution' cannot be directly changed after simulation starts. Use "
            "'update_attr_array()'."
        )
        if value is not None:
            value = np.asanyarray(value, dtype=np.float32)
            if value.ndim not in (
                self._ATTR_ARRAY_NDIM["link_restitution"],
                self._ATTR_ARRAY_NDIM["link_restitution"] + 1,
            ):
                raise ValueError(
                    "'link_restitution' must have a number of dimensions of "
                    f"{self._ATTR_ARRAY_NDIM['link_restitution']} or "
                    f"{self._ATTR_ARRAY_NDIM['link_restitution'] + 1}"
                )
        self._link_restitution = value

    @property
    def link_linear_damping(self):
        """ """
        return self._link_linear_damping

    @link_linear_damping.setter
    def link_linear_damping(self, value):
        """ """
        assert not self._attr_array_locked["link_linear_damping"], (
            "'link_linear_damping' cannot be directly changed after simulation starts. Use "
            "'update_attr_array()'."
        )
        if value is not None:
            value = np.asanyarray(value, dtype=np.float32)
            if value.ndim not in (
                self._ATTR_ARRAY_NDIM["link_linear_damping"],
                self._ATTR_ARRAY_NDIM["link_linear_damping"] + 1,
            ):
                raise ValueError(
                    "'link_linear_damping' must have a number of dimensions of "
                    f"{self._ATTR_ARRAY_NDIM['link_linear_damping']} or "
                    f"{self._ATTR_ARRAY_NDIM['link_linear_damping'] + 1}"
                )
        self._link_linear_damping = value

    @property
    def link_angular_damping(self):
        """ """
        return self._link_angular_damping

    @link_angular_damping.setter
    def link_angular_damping(self, value):
        """ """
        assert not self._attr_array_locked["link_angular_damping"], (
            "'link_angular_damping' cannot be directly changed after simulation starts. Use "
            "'update_attr_array()'."
        )
        if value is not None:
            value = np.asanyarray(value, dtype=np.float32)
            if value.ndim not in (
                self._ATTR_ARRAY_NDIM["link_angular_damping"],
                self._ATTR_ARRAY_NDIM["link_angular_damping"] + 1,
            ):
                raise ValueError(
                    "'link_angular_damping' must have a number of dimensions of "
                    f"{self._ATTR_ARRAY_NDIM['link_angular_damping']} or "
                    f"{self._ATTR_ARRAY_NDIM['link_angular_damping'] + 1}"
                )
        self._link_angular_damping = value

    @property
    def link_color(self):
        """ """
        return self._link_color

    @link_color.setter
    def link_color(self, value):
        """ """
        assert not self._attr_array_locked["link_color"], (
            "'link_color' cannot be directly changed after simulation starts. Use "
            "'update_attr_array()'."
        )
        if value is not None:
            value = np.asanyarray(value, dtype=np.float32)
            if value.ndim not in (
                self._ATTR_ARRAY_NDIM["link_color"],
                self._ATTR_ARRAY_NDIM["link_color"] + 1,
            ):
                raise ValueError(
                    "'link_color' must have a number of dimensions of "
                    f"{self._ATTR_ARRAY_NDIM['link_color']} or "
                    f"{self._ATTR_ARRAY_NDIM['link_color'] + 1}"
                )
        self._link_color = value

    @property
    def link_segmentation_id(self):
        """ """
        return self._link_segmentation_id

    @link_segmentation_id.setter
    def link_segmentation_id(self, value):
        """ """
        assert not self._attr_array_locked["link_segmentation_id"], (
            "'link_segmentation_id' cannot be directly changed after simulation starts. Use "
            "'update_attr_array()'."
        )
        if value is not None:
            value = np.asanyarray(value, dtype=np.int64)
            if value.ndim not in (
                self._ATTR_ARRAY_NDIM["link_segmentation_id"],
                self._ATTR_ARRAY_NDIM["link_segmentation_id"] + 1,
            ):
                raise ValueError(
                    "'link_segmentation_id' must have a number of dimensions of "
                    f"{self._ATTR_ARRAY_NDIM['link_segmentation_id']} or "
                    f"{self._ATTR_ARRAY_NDIM['link_segmentation_id'] + 1}"
                )
        self._link_segmentation_id = value

    @property
    def dof_has_limits(self):
        """ """
        return self._dof_has_limits

    @dof_has_limits.setter
    def dof_has_limits(self, value):
        """ """
        assert not self._attr_array_locked["dof_has_limits"], (
            "'dof_has_limits' cannot be directly changed after simulation starts. Use "
            "'update_attr_array()'."
        )
        if value is not None:
            value = np.asanyarray(value, dtype=np.bool_)
            if value.ndim not in (
                self._ATTR_ARRAY_NDIM["dof_has_limits"],
                self._ATTR_ARRAY_NDIM["dof_has_limits"] + 1,
            ):
                raise ValueError(
                    "'dof_has_limits' must have a number of dimensions of "
                    f"{self._ATTR_ARRAY_NDIM['dof_has_limits']} or "
                    f"{self._ATTR_ARRAY_NDIM['dof_has_limits'] + 1}"
                )
        self._dof_has_limits = value

    @property
    def dof_lower_limit(self):
        """ """
        return self._dof_lower_limit

    @dof_lower_limit.setter
    def dof_lower_limit(self, value):
        """ """
        assert not self._attr_array_locked["dof_lower_limit"], (
            "'dof_lower_limit' cannot be directly changed after simulation starts. Use "
            "'update_attr_array()'."
        )
        if value is not None:
            value = np.asanyarray(value, dtype=np.float32)
            if value.ndim not in (
                self._ATTR_ARRAY_NDIM["dof_lower_limit"],
                self._ATTR_ARRAY_NDIM["dof_lower_limit"] + 1,
            ):
                raise ValueError(
                    "'dof_lower_limit' must have a number of dimensions of "
                    f"{self._ATTR_ARRAY_NDIM['dof_lower_limit']} or "
                    f"{self._ATTR_ARRAY_NDIM['dof_lower_limit'] + 1}"
                )
        self._dof_lower_limit = value

    @property
    def dof_upper_limit(self):
        """ """
        return self._dof_upper_limit

    @dof_upper_limit.setter
    def dof_upper_limit(self, value):
        """ """
        assert not self._attr_array_locked["dof_upper_limit"], (
            "'dof_upper_limit' cannot be directly changed after simulation starts. Use "
            "'update_attr_array()'."
        )
        if value is not None:
            value = np.asanyarray(value, dtype=np.float32)
            if value.ndim not in (
                self._ATTR_ARRAY_NDIM["dof_upper_limit"],
                self._ATTR_ARRAY_NDIM["dof_upper_limit"] + 1,
            ):
                raise ValueError(
                    "'dof_upper_limit' must have a number of dimensions of "
                    f"{self._ATTR_ARRAY_NDIM['dof_upper_limit']} or "
                    f"{self._ATTR_ARRAY_NDIM['dof_upper_limit'] + 1}"
                )
        self._dof_upper_limit = value

    @property
    def dof_control_mode(self):
        """ """
        return self._dof_control_mode

    @dof_control_mode.setter
    def dof_control_mode(self, value):
        """ """
        assert not self._attr_array_locked["dof_control_mode"], (
            "'dof_control_mode' cannot be directly changed after simulation starts. Use "
            "'update_attr_array()'."
        )
        if value is not None:
            value = np.asanyarray(value, dtype=np.int64)
            if value.ndim not in (
                self._ATTR_ARRAY_NDIM["dof_control_mode"],
                self._ATTR_ARRAY_NDIM["dof_control_mode"] + 1,
            ):
                raise ValueError(
                    "'dof_control_mode' must have a number of dimensions of "
                    f"{self._ATTR_ARRAY_NDIM['dof_control_mode']} or "
                    f"{self._ATTR_ARRAY_NDIM['dof_control_mode'] + 1}"
                )
        self._dof_control_mode = value

    @property
    def dof_max_force(self):
        """ """
        return self._dof_max_force

    @dof_max_force.setter
    def dof_max_force(self, value):
        """ """
        assert not self._attr_array_locked["dof_max_force"], (
            "'dof_max_force' cannot be directly changed after simulation starts. Use "
            "'update_attr_array()'."
        )
        if value is not None:
            value = np.asanyarray(value, dtype=np.float32)
            if value.ndim not in (
                self._ATTR_ARRAY_NDIM["dof_max_force"],
                self._ATTR_ARRAY_NDIM["dof_max_force"] + 1,
            ):
                raise ValueError(
                    "'dof_max_force' must have a number of dimensions of "
                    f"{self._ATTR_ARRAY_NDIM['dof_max_force']} or "
                    f"{self._ATTR_ARRAY_NDIM['dof_max_force'] + 1}"
                )
        self._dof_max_force = value

    @property
    def dof_max_velocity(self):
        """ """
        return self._dof_max_velocity

    @dof_max_velocity.setter
    def dof_max_velocity(self, value):
        """ """
        assert not self._attr_array_locked["dof_max_velocity"], (
            "'dof_max_velocity' cannot be directly changed after simulation starts. Use "
            "'update_attr_array()'."
        )
        if value is not None:
            value = np.asanyarray(value, dtype=np.float32)
            if value.ndim not in (
                self._ATTR_ARRAY_NDIM["dof_max_velocity"],
                self._ATTR_ARRAY_NDIM["dof_max_velocity"] + 1,
            ):
                raise ValueError(
                    "'dof_max_velocity' must have a number of dimensions of "
                    f"{self._ATTR_ARRAY_NDIM['dof_max_velocity']} or "
                    f"{self._ATTR_ARRAY_NDIM['dof_max_velocity'] + 1}"
                )
        self._dof_max_velocity = value

    @property
    def dof_position_gain(self):
        """ """
        return self._dof_position_gain

    @dof_position_gain.setter
    def dof_position_gain(self, value):
        """ """
        assert not self._attr_array_locked["dof_position_gain"], (
            "'dof_position_gain' cannot be directly changed after simulation starts. Use "
            "'update_attr_array()'."
        )
        if value is not None:
            value = np.asanyarray(value, dtype=np.float32)
            if value.ndim not in (
                self._ATTR_ARRAY_NDIM["dof_position_gain"],
                self._ATTR_ARRAY_NDIM["dof_position_gain"] + 1,
            ):
                raise ValueError(
                    "'dof_position_gain' must have a number of dimensions of "
                    f"{self._ATTR_ARRAY_NDIM['dof_position_gain']} or "
                    f"{self._ATTR_ARRAY_NDIM['dof_position_gain'] + 1}"
                )
        self._dof_position_gain = value

    @property
    def dof_velocity_gain(self):
        """ """
        return self._dof_velocity_gain

    @dof_velocity_gain.setter
    def dof_velocity_gain(self, value):
        """ """
        assert not self._attr_array_locked["dof_velocity_gain"], (
            "'dof_velocity_gain' cannot be directly changed after simulation starts. Use "
            "'update_attr_array()'."
        )
        if value is not None:
            value = np.asanyarray(value, dtype=np.float32)
            if value.ndim not in (
                self._ATTR_ARRAY_NDIM["dof_velocity_gain"],
                self._ATTR_ARRAY_NDIM["dof_velocity_gain"] + 1,
            ):
                raise ValueError(
                    "'dof_velocity_gain' must have a number of dimensions of "
                    f"{self._ATTR_ARRAY_NDIM['dof_velocity_gain']} or "
                    f"{self._ATTR_ARRAY_NDIM['dof_velocity_gain'] + 1}"
                )
        self._dof_velocity_gain = value

    @property
    def dof_armature(self):
        """ """
        return self._dof_armature

    @dof_armature.setter
    def dof_armature(self, value):
        """ """
        assert not self._attr_array_locked["dof_armature"], (
            "'dof_armature' cannot be directly changed after simulation starts. Use "
            "'update_attr_array()'."
        )
        if value is not None:
            value = np.asanyarray(value, dtype=np.float32)
            if value.ndim not in (
                self._ATTR_ARRAY_NDIM["dof_armature"],
                self._ATTR_ARRAY_NDIM["dof_armature"] + 1,
            ):
                raise ValueError(
                    "'dof_armature' must have a number of dimensions of "
                    f"{self._ATTR_ARRAY_NDIM['dof_armature']} or "
                    f"{self._ATTR_ARRAY_NDIM['dof_armature'] + 1}"
                )
        self._dof_armature = value

    @property
    def dof_target_position(self):
        """ """
        return self._dof_target_position

    @dof_target_position.setter
    def dof_target_position(self, value):
        """ """
        if value is not None:
            value = torch.as_tensor(value, dtype=torch.float32, device=self.device)
            if value.ndim not in (
                self._ATTR_TENSOR_NDIM["dof_target_position"],
                self._ATTR_TENSOR_NDIM["dof_target_position"] + 1,
            ):
                raise ValueError(
                    "'dof_target_position' must have a number of dimensions of "
                    f"{self._ATTR_TENSOR_NDIM['dof_target_position']} or "
                    f"{self._ATTR_TENSOR_NDIM['dof_target_position'] + 1}"
                )
        self._dof_target_position = value

    @property
    def dof_target_velocity(self):
        """ """
        return self._dof_target_velocity

    @dof_target_velocity.setter
    def dof_target_velocity(self, value):
        """ """
        if value is not None:
            value = torch.as_tensor(value, dtype=torch.float32, device=self.device)
            if value.ndim not in (
                self._ATTR_TENSOR_NDIM["dof_target_velocity"],
                self._ATTR_TENSOR_NDIM["dof_target_velocity"] + 1,
            ):
                raise ValueError(
                    "'dof_target_velocity' must have a number of dimensions of "
                    f"{self._ATTR_TENSOR_NDIM['dof_target_velocity']} or "
                    f"{self._ATTR_TENSOR_NDIM['dof_target_velocity'] + 1}"
                )
        self._dof_target_velocity = value

    @property
    def dof_force(self):
        """ """
        return self._dof_force

    @dof_force.setter
    def dof_force(self, value):
        """ """
        if value is not None:
            value = torch.as_tensor(value, dtype=torch.float32, device=self.device)
            if value.ndim not in (
                self._ATTR_TENSOR_NDIM["dof_force"],
                self._ATTR_TENSOR_NDIM["dof_force"] + 1,
            ):
                raise ValueError(
                    "'dof_force' must have a number of dimensions of "
                    f"{self._ATTR_TENSOR_NDIM['dof_force']} or "
                    f"{self._ATTR_TENSOR_NDIM['dof_force'] + 1}"
                )
        self._dof_force = value

    @property
    def env_ids_reset_base_state(self):
        """ """
        return self._env_ids_reset_base_state

    @env_ids_reset_base_state.setter
    def env_ids_reset_base_state(self, value):
        """ """
        if value is not None:
            value = torch.as_tensor(value, dtype=torch.int64, device=self.device)
            if value.ndim != 1:
                raise ValueError("'env_ids_reset_base_state' must have a number of dimensions of 1")
        self._env_ids_reset_base_state = value

    @property
    def env_ids_reset_dof_state(self):
        """ """
        return self._env_ids_reset_dof_state

    @env_ids_reset_dof_state.setter
    def env_ids_reset_dof_state(self, value):
        """ """
        if value is not None:
            value = torch.as_tensor(value, dtype=torch.int64, device=self.device)
            if value.ndim != 1:
                raise ValueError("'env_ids_reset_dof_state' must have a number of dimensions of 1")
        self._env_ids_reset_dof_state = value

    @property
    def dof_state(self):
        """ """
        if self._dof_state is None and self._callback_collect_dof_state is not None:
            self._callback_collect_dof_state(self)
        return self._dof_state

    @dof_state.setter
    def dof_state(self, value):
        """ """
        if value is not None:
            value = torch.as_tensor(value, dtype=torch.float32, device=self.device)
        self._dof_state = value

    @property
    def link_state(self):
        """ """
        if self._link_state is None and self._callback_collect_link_state is not None:
            self._callback_collect_link_state(self)
        return self._link_state

    @link_state.setter
    def link_state(self, value):
        """ """
        if value is not None:
            value = torch.as_tensor(value, dtype=torch.float32, device=self.device)
        self._link_state = value

    @property
    def contact_id(self):
        """ """
        return self._contact_id

    @contact_id.setter
    def contact_id(self, value):
        """ """
        if value is not None:
            value = np.asanyarray(value, dtype=np.int64)
        self._contact_id = value

    def get_attr_array(self, attr, idx):
        """ """
        return self._get_attr(attr, self._ATTR_ARRAY_NDIM[attr], idx)

    def get_attr_tensor(self, attr, idx):
        """ """
        return self._get_attr(attr, self._ATTR_TENSOR_NDIM[attr], idx)

    def _get_attr(self, attr, ndim, idx):
        """ """
        array = getattr(self, attr)
        if array.ndim == ndim:
            return array
        if array.ndim == ndim + 1:
            return array[idx]

    def lock_attr_array(self):
        """ """
        for k in self._attr_array_locked:
            if not self._attr_array_locked[k]:
                self._attr_array_locked[k] = True
            if getattr(self, k) is not None:
                getattr(self, k).flags.writeable = False

    def update_attr_array(self, attr, env_ids, value):
        """ """
        if getattr(self, attr).ndim != self._ATTR_ARRAY_NDIM[attr] + 1:
            raise ValueError(
                f"'{attr}' can only be updated when a per-env specification (ndim: "
                f"{self._ATTR_ARRAY_NDIM[attr] + 1}) is used"
            )
        if len(env_ids) == 0:
            return

        env_ids_np = env_ids.cpu().numpy()

        with self._make_attr_array_writeable(attr):
            getattr(self, attr)[env_ids_np] = value

        if not self._attr_array_dirty_flag[attr]:
            self._attr_array_dirty_flag[attr] = True
        try:
            self._attr_array_dirty_mask[attr][env_ids_np] = True
        except KeyError:
            self._attr_array_dirty_mask[attr] = np.zeros(len(getattr(self, attr)), dtype=bool)
            self._attr_array_dirty_mask[attr][env_ids_np] = True

        if self._attr_array_default_flag[attr]:
            self._attr_array_default_flag[attr] = False

    @contextmanager
    def _make_attr_array_writeable(self, attr):
        """ """
        try:
            getattr(self, attr).flags.writeable = True
            yield
        finally:
            getattr(self, attr).flags.writeable = False

    def set_callback_collect_dof_state(self, callback):
        """ """
        self._callback_collect_dof_state = callback

    def set_callback_collect_link_state(self, callback):
        """ """
        self._callback_collect_link_state = callback



# Copyright (c) 2021, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# Licensed under the MIT License [see LICENSE for details].

import gym
import abc

from easysim.scene import Scene
from easysim.simulators.registration import make


class SimulatorEnv(gym.Env, abc.ABC):
    """ """

    def __init__(self, cfg, **kwargs):
        """ """
        self._cfg = cfg

        self._scene = Scene()

        self.init(**kwargs)

        self._simulator = make(self.cfg.SIM.SIMULATOR, cfg=self.cfg.SIM, scene=self.scene)

    @property
    def cfg(self):
        """ """
        return self._cfg

    @property
    def scene(self):
        """ """
        return self._scene

    @property
    def simulator(self):
        """ """
        return self._simulator

    @abc.abstractmethod
    def init(self, **kwargs):
        """ """

    def reset(self, env_ids=None, **kwargs):
        """ """
        self.pre_reset(env_ids, **kwargs)

        self._simulator.reset(env_ids)

        observation = self.post_reset(env_ids, **kwargs)

        return observation

    @abc.abstractmethod
    def pre_reset(self, env_ids, **kwargs):
        """ """

    @abc.abstractmethod
    def post_reset(self, env_ids, **kwargs):
        """ """

    def step(self, action):
        """ """
        self.pre_step(action)

        self._simulator.step()

        observation, reward, done, info = self.post_step(action)

        return observation, reward, done, info

    @abc.abstractmethod
    def pre_step(self, action):
        """ """

    @abc.abstractmethod
    def post_step(self, action):
        """ """

    @property
    def contact(self):
        """ """
        return self._simulator.contact

    def close(self):
        """ """
        self._simulator.close()


class SimulatorWrapper(gym.Wrapper):
    """ """

    def reset(self, env_ids=None, **kwargs):
        """ """
        return self.env.reset(env_ids=env_ids, **kwargs)

    def step(self, action):
        """ """
        return self.env.step(action)

    def close(self):
        """ """
        return self.env.close()



# Copyright (c) 2021, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# Licensed under the MIT License [see LICENSE for details].

from easysim.body import Body
from easysim.camera import Camera


class Scene:
    """ """

    def __init__(self):
        # A `list` is adopted here rather than a `set`, because for some simulators, the simulation
        # outcome is variant to the loading order of bodies. Using a `list` can enforce a
        # deterministic and controllable order.
        self._bodies = list()

        self._name_to_body = {}

        self._cameras = set()

        self._name_to_camera = {}

        self._init_callback()

    @property
    def bodies(self):
        """ """
        return self._bodies

    @property
    def cameras(self):
        """ """
        return self._cameras

    def _init_callback(self):
        """ """
        self._callback_add_camera = None
        self._callback_remove_camera = None

    def add_body(self, body):
        """ """
        if not isinstance(body, Body):
            raise TypeError("body must be a Body")
        if body in self.bodies:
            raise ValueError("body already in scene")
        if body.name is None:
            raise ValueError("body.name must not be None")
        if body.name in self._name_to_body:
            raise ValueError(f"Cannot add body with duplicated name: '{body.name}'")
        self.bodies.append(body)

        self._name_to_body[body.name] = body

    def remove_body(self, body):
        """ """
        if body not in self.bodies:
            raise ValueError("body not in the scene")
        self.bodies.remove(body)

        del self._name_to_body[body.name]

    def get_body(self, name):
        """ """
        if name not in self._name_to_body:
            raise ValueError(f"Non-existent body name: '{name}'")
        return self._name_to_body[name]

    def add_camera(self, camera):
        """ """
        if not isinstance(camera, Camera):
            raise TypeError("camera must be a Camera")
        if camera in self.cameras:
            raise ValueError("camera already in scene")
        if camera.name is None:
            raise ValueError("camera.name must not be None")
        if camera.name in self._name_to_camera:
            raise ValueError(f"Cannot add camera with duplicated name: '{camera.name}")
        self.cameras.add(camera)

        self._name_to_camera[camera.name] = camera

        if self._callback_add_camera is not None:
            self._callback_add_camera(camera)

    def remove_camera(self, camera):
        """ """
        if camera not in self.cameras:
            raise ValueError("camera not in the scene")
        self.cameras.remove(camera)

        del self._name_to_camera[camera.name]

        if self._callback_remove_camera is not None:
            self._callback_remove_camera(camera)

    def get_camera(self, name):
        """ """
        if name not in self._name_to_camera:
            raise ValueError(f"Non-existent camera name: '{name}'")
        return self._name_to_camera[name]

    def set_callback_add_camera(self, callback):
        """ """
        self._callback_add_camera = callback

    def set_callback_remove_camera(self, callback):
        """ """
        self._callback_remove_camera = callback



# Copyright (c) 2021, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# Licensed under the MIT License [see LICENSE for details].

"""EasySim package."""


# NOTE (roflaherty): This is inspired by how matplotlib does creates its version value.
# https://github.com/matplotlib/matplotlib/blob/master/lib/matplotlib/__init__.py#L161
def _get_version() -> str:
    """Return the version string used for __version__."""
    # Standard Library
    import pathlib

    root = pathlib.Path(__file__).resolve().parent.parent.parent
    if (root / ".git").exists() and not (root / ".git/shallow").exists():
        # Third Party
        import setuptools_scm

        this_version: str
        # See the `setuptools_scm` documentation for the description of the schemes used below.
        # https://pypi.org/project/setuptools-scm/
        # NOTE: If these values are updated, they need to be also updated in `pyproject.toml`.
        this_version = setuptools_scm.get_version(
            root=root,
            version_scheme="no-guess-dev",
            local_scheme="dirty-tag",
        )
    else:  # Get the version from the _version.py setuptools_scm file.
        try:
            # Standard Library
            from importlib.metadata import version
        except ModuleNotFoundError:
            # NOTE: `importlib.resources` is part of the standard library in Python 3.9.
            # `importlib_metadata` is the back ported library for older versions of python.
            # Third Party
            from importlib_metadata import version  # type: ignore[no-redef]

        this_version = version("easysim")

    return this_version


# Set `__version__` attribute
__version__ = _get_version()

# Remove `_get_version` so it is not added as an attribute
del _get_version

from easysim.simulator_env import SimulatorEnv, SimulatorWrapper
from easysim.body import Body
from easysim.camera import Camera
from easysim.constants import GeometryType, DoFControlMode, MeshNormalMode
from easysim.config import cfg, get_cfg
from easysim.cmd import get_config_from_args

__all__ = [
    "SimulatorEnv",
    "SimulatorWrapper",
    "Body",
    "Camera",
    "GeometryType",
    "DoFControlMode",
    "MeshNormalMode",
    "cfg",
    "get_cfg",
    "get_config_from_args",
]



# Copyright (c) 2021, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# Licensed under the MIT License [see LICENSE for details].

import pybullet
import pybullet_utils.bullet_client as bullet_client
import pybullet_data
import pkgutil
import numpy as np
import time

from contextlib import contextmanager

from easysim.simulators.simulator import Simulator
from easysim.constants import GeometryType, DoFControlMode
from easysim.contact import create_contact_array


class Bullet(Simulator):
    """Bullet simulator."""

    _ATTR_LINK_DYNAMICS = (
        "link_lateral_friction",
        "link_spinning_friction",
        "link_rolling_friction",
        "link_restitution",
        "link_linear_damping",
        "link_angular_damping",
    )
    _ATTR_DOF_DYNAMICS = (
        "dof_lower_limit",
        "dof_upper_limit",
    )
    _ATTR_PROJECTION_MATRIX = ("width", "height", "vertical_fov", "near", "far")
    _ATTR_VIEW_MATRIX = ("position", "target", "up_vector", "orientation")
    _DOF_CONTROL_MODE_MAP = {
        DoFControlMode.POSITION_CONTROL: pybullet.POSITION_CONTROL,
        DoFControlMode.VELOCITY_CONTROL: pybullet.VELOCITY_CONTROL,
        DoFControlMode.TORQUE_CONTROL: pybullet.TORQUE_CONTROL,
    }

    def __init__(self, cfg, scene):
        """ """
        super().__init__(cfg, scene)

        if not self._cfg.DRAW_VIEWER_AXES:
            raise ValueError("DRAW_VIEWER_AXES must be True for Bullet")
        if self._cfg.NUM_ENVS != 1:
            raise ValueError("NUM_ENVS must be 1 for Bullet")
        if self._cfg.SIM_DEVICE != "cpu":
            raise ValueError("SIM_DEVICE must be 'cpu' for Bullet")

        self._connected = False
        self._last_frame_time = 0.0

    def reset(self, env_ids):
        """ """
        if not self._connected:
            if self._cfg.RENDER:
                self._p = bullet_client.BulletClient(connection_mode=pybullet.GUI)
            else:
                self._p = bullet_client.BulletClient(connection_mode=pybullet.DIRECT)
            self._p.setAdditionalSearchPath(pybullet_data.getDataPath())

            self._plugins = {}

            if self._cfg.BULLET.USE_EGL:
                if self._cfg.RENDER:
                    raise ValueError("USE_EGL can only be True when RENDER is set to False")
                egl = pkgutil.get_loader("eglRenderer")
                self._plugins["egl_renderer"] = self._p.loadPlugin(
                    egl.get_filename(), "_eglRendererPlugin"
                )

        with self._disable_cov_rendering():
            self._p.resetSimulation()
            self._p.setGravity(*self._cfg.GRAVITY)
            if self._cfg.USE_DEFAULT_STEP_PARAMS:
                sim_params = self._p.getPhysicsEngineParameters()
                self._cfg.TIME_STEP = sim_params["fixedTimeStep"]
                self._cfg.SUBSTEPS = max(sim_params["numSubSteps"], 1)
            else:
                self._p.setPhysicsEngineParameter(
                    fixedTimeStep=self._cfg.TIME_STEP, numSubSteps=self._cfg.SUBSTEPS
                )
            self._p.setPhysicsEngineParameter(deterministicOverlappingPairs=1)

            if self._cfg.GROUND_PLANE.LOAD:
                self._body_id_ground_plane = self._load_ground_plane()

            self._scene_cache = type(self._scene)()

            self._body_ids = {}
            self._dof_indices = {}
            self._num_links = {}

            for body in self._scene.bodies:
                self._load_body(body)
                self._cache_body_and_set_control_and_props(body)
                self._set_callback_body(body)

            self._projection_matrix = {}
            self._view_matrix = {}
            self._image_cache = {}

            for camera in self._scene.cameras:
                self._load_camera(camera)
                self._set_callback_camera(camera)

            if not self._connected:
                self._set_callback_scene()

            if (
                self._cfg.RENDER
                and self._cfg.INIT_VIEWER_CAMERA_POSITION != (None, None, None)
                and self._cfg.INIT_VIEWER_CAMERA_TARGET != (None, None, None)
            ):
                self._set_viewer_camera_pose(
                    self._cfg.INIT_VIEWER_CAMERA_POSITION, self._cfg.INIT_VIEWER_CAMERA_TARGET
                )

        if not self._connected:
            self._connected = True

        self._clear_state()
        self._clear_image()
        self._contact = None

    @contextmanager
    def _disable_cov_rendering(self):
        """ """
        try:
            if self._cfg.RENDER:
                self._p.configureDebugVisualizer(pybullet.COV_ENABLE_RENDERING, 0)
            yield
        finally:
            if self._cfg.RENDER:
                self._p.configureDebugVisualizer(pybullet.COV_ENABLE_RENDERING, 1)

    def _load_ground_plane(self):
        """ """
        kwargs = {}
        kwargs["basePosition"] = (0.0, 0.0, -self._cfg.GROUND_PLANE.DISTANCE)
        return self._p.loadURDF("plane_implicit.urdf", **kwargs)

    def _load_body(self, body):
        """ """
        if body.env_ids_load is not None:
            if np.array_equal(body.env_ids_load.cpu(), []):
                return
            elif not np.array_equal(body.env_ids_load.cpu(), [0]):
                raise ValueError(
                    f"For Bullet, 'env_ids_load' must be either None, [] or [0]: '{body.name}'"
                )

        for attr in ("vhacd_enabled", "vhacd_params", "mesh_normal_mode"):
            if getattr(body, attr) is not None:
                raise ValueError(f"'{attr}' is not supported in Bullet: '{body.name}'")
        kwargs = {}
        if body.use_self_collision is not None and body.use_self_collision:
            kwargs["flags"] = pybullet.URDF_USE_SELF_COLLISION
        if body.geometry_type is None:
            raise ValueError(f"For Bullet, 'geometry_type' must not be None: '{body.name}'")
        if body.geometry_type not in (GeometryType.URDF, GeometryType.SPHERE, GeometryType.BOX):
            raise ValueError(
                f"For Bullet, 'geometry_type' only supports URDF, SPHERE, and BOX: '{body.name}'"
            )
        if body.geometry_type == GeometryType.URDF:
            for attr in ("sphere_radius", "box_half_extent"):
                if getattr(body, attr) is not None:
                    raise ValueError(f"'{attr}' must be None for geometry type URDF: '{body.name}'")
            if body.use_fixed_base is not None:
                kwargs["useFixedBase"] = body.use_fixed_base
            if body.scale is not None:
                kwargs["globalScaling"] = body.get_attr_array("scale", 0)
                if body.attr_array_dirty_flag["scale"]:
                    body.attr_array_dirty_flag["scale"] = False
            self._body_ids[body.name] = self._p.loadURDF(body.urdf_file, **kwargs)
        else:
            for attr in ("scale",):
                if getattr(body, attr) is not None:
                    geometry_type = [
                        x
                        for x in dir(GeometryType)
                        if not x.startswith("__") and getattr(GeometryType, x) == body.geometry_type
                    ][0]
                    raise ValueError(
                        f"For Bullet, '{attr}' is not supported for geometry type {geometry_type}:"
                        f" '{body.name}'"
                    )
            kwargs_visual = {}
            kwargs_collision = {}
            if body.geometry_type == GeometryType.SPHERE:
                for attr in ("urdf_file", "box_half_extent"):
                    if getattr(body, attr) is not None:
                        raise ValueError(
                            f"'{attr}' must be None for geometry type SPHERE: '{body.name}'"
                        )
                if body.sphere_radius is not None:
                    kwargs_visual["radius"] = body.sphere_radius
                    kwargs_collision["radius"] = body.sphere_radius
                kwargs["baseVisualShapeIndex"] = self._p.createVisualShape(
                    pybullet.GEOM_SPHERE, **kwargs_visual
                )
                kwargs["baseCollisionShapeIndex"] = self._p.createCollisionShape(
                    pybullet.GEOM_SPHERE, **kwargs_collision
                )
            if body.geometry_type == GeometryType.BOX:
                for attr in ("urdf_file", "sphere_radius"):
                    if getattr(body, attr) is not None:
                        raise ValueError(
                            f"'{attr}' must be None for geometry type BOX: '{body.name}'"
                        )
                if body.box_half_extent is not None:
                    kwargs_visual["halfExtents"] = body.box_half_extent
                    kwargs_collision["halfExtents"] = body.box_half_extent
                kwargs["baseVisualShapeIndex"] = self._p.createVisualShape(
                    pybullet.GEOM_BOX, **kwargs_visual
                )
                kwargs["baseCollisionShapeIndex"] = self._p.createCollisionShape(
                    pybullet.GEOM_BOX, **kwargs_collision
                )
            if body.use_fixed_base is not None and body.use_fixed_base:
                kwargs["baseMass"] = 0.0
            self._body_ids[body.name] = self._p.createMultiBody(**kwargs)

        dof_indices = []
        for j in range(self._p.getNumJoints(self._body_ids[body.name])):
            joint_info = self._p.getJointInfo(self._body_ids[body.name], j)
            if joint_info[2] != pybullet.JOINT_FIXED:
                dof_indices.append(j)
        self._dof_indices[body.name] = np.asanyarray(dof_indices, dtype=np.int64)

        self._num_links[body.name] = self._p.getNumJoints(self._body_ids[body.name]) + 1

        # Reset base state.
        if body.initial_base_position is not None:
            self._reset_base_position(body)
        if body.initial_base_velocity is not None:
            self._reset_base_velocity(body)

        # Reset DoF state.
        if len(self._dof_indices[body.name]) == 0:
            for attr in ("initial_dof_position", "initial_dof_velocity"):
                if getattr(body, attr) is not None:
                    raise ValueError(f"'{attr}' must be None for body with 0 DoF: '{body.name}'")
        if body.initial_dof_position is not None:
            self._reset_dof_state(body)
        elif body.initial_dof_velocity is not None:
            raise ValueError(
                "For Bullet, cannot reset 'initial_dof_velocity' without resetting "
                f"'initial_dof_position': '{body.name}'"
            )

        body.contact_id = [self._body_ids[body.name]]

    def _reset_base_position(self, body):
        """ """
        if body.initial_base_position.ndim == 1:
            self._p.resetBasePositionAndOrientation(
                self._body_ids[body.name],
                body.initial_base_position[:3],
                body.initial_base_position[3:],
            )
        if body.initial_base_position.ndim == 2:
            self._p.resetBasePositionAndOrientation(
                self._body_ids[body.name],
                body.initial_base_position[0, :3],
                body.initial_base_position[0, 3:],
            )

    def _reset_base_velocity(self, body):
        """ """
        kwargs = {}
        if body.initial_base_velocity.ndim == 1:
            kwargs["linearVelocity"] = body.initial_base_velocity[:3]
            kwargs["angularVelocity"] = body.initial_base_velocity[3:]
        if body.initial_base_velocity.ndim == 2:
            kwargs["linearVelocity"] = body.initial_base_velocity[0, :3]
            kwargs["angularVelocity"] = body.initial_base_velocity[0, 3:]
        self._p.resetBaseVelocity(self._body_ids[body.name], **kwargs)

    def _reset_dof_state(self, body):
        """ """
        for i, j in enumerate(self._dof_indices[body.name]):
            kwargs = {}
            if body.initial_dof_velocity is not None:
                if body.initial_dof_velocity.ndim == 1:
                    kwargs["targetVelocity"] = body.initial_dof_velocity[i]
                if body.initial_dof_velocity.ndim == 2:
                    kwargs["targetVelocity"] = body.initial_dof_velocity[0, i]
            if body.initial_dof_position.ndim == 1:
                self._p.resetJointState(
                    self._body_ids[body.name], j, body.initial_dof_position[i], **kwargs
                )
            if body.initial_dof_position.ndim == 2:
                self._p.resetJointState(
                    self._body_ids[body.name], j, body.initial_dof_position[0, i], **kwargs
                )

    def _cache_body_and_set_control_and_props(self, body):
        """ """
        x = type(body)()
        x.name = body.name
        self._scene_cache.add_body(x)

        if body.env_ids_load is not None and len(body.env_ids_load) == 0:
            body.lock_attr_array()
            return

        for attr in ("link_segmentation_id", "dof_has_limits", "dof_armature"):
            if getattr(body, attr) is not None:
                raise ValueError(f"'{attr}' is not supported in Bullet: '{body.name}'")

        if len(self._dof_indices[body.name]) == 0:
            for attr in ("dof_lower_limit", "dof_upper_limit", "dof_control_mode"):
                if getattr(body, attr) is not None:
                    raise ValueError(f"'{attr}' must be None for body with 0 DoF: '{body.name}'")

        if body.dof_control_mode is not None:
            if (
                body.dof_control_mode.ndim == 0
                and body.dof_control_mode
                not in (
                    DoFControlMode.POSITION_CONTROL,
                    DoFControlMode.VELOCITY_CONTROL,
                    DoFControlMode.TORQUE_CONTROL,
                )
                or body.dof_control_mode.ndim == 1
                and any(
                    y
                    not in (
                        DoFControlMode.POSITION_CONTROL,
                        DoFControlMode.VELOCITY_CONTROL,
                        DoFControlMode.TORQUE_CONTROL,
                    )
                    for y in body.dof_control_mode
                )
            ):
                raise ValueError(
                    "For Bullet, 'dof_control_mode' only supports POSITION_CONTROL, "
                    f"VELOCITY_CONTROL, and TORQUE_CONTROL: '{body.name}'"
                )

            if (
                body.dof_control_mode.ndim == 0
                and body.dof_control_mode == DoFControlMode.TORQUE_CONTROL
            ):
                self._p.setJointMotorControlArray(
                    self._body_ids[body.name],
                    self._dof_indices[body.name],
                    pybullet.VELOCITY_CONTROL,
                    forces=[0] * len(self._dof_indices[body.name]),
                )
            if (
                body.dof_control_mode.ndim == 1
                and DoFControlMode.TORQUE_CONTROL in body.dof_control_mode
            ):
                self._p.setJointMotorControlArray(
                    self._body_ids[body.name],
                    self._dof_indices[body.name][
                        body.dof_control_mode == DoFControlMode.TORQUE_CONTROL
                    ],
                    pybullet.VELOCITY_CONTROL,
                    forces=[0]
                    * len(
                        self._dof_indices[body.name][
                            body.dof_control_mode == DoFControlMode.TORQUE_CONTROL
                        ]
                    ),
                )
        elif len(self._dof_indices[body.name]) > 0:
            raise ValueError(
                f"For Bullet, 'dof_control_mode' is required for body with DoF > 0: '{body.name}'"
            )

        if body.link_collision_filter is not None:
            self._set_link_collision_filter(body)

        if any(
            not body.attr_array_default_flag[x] and getattr(body, x) is not None
            for x in self._ATTR_LINK_DYNAMICS
        ):
            self._set_link_dynamics(body)

        if not body.attr_array_default_flag["link_color"] and body.link_color is not None:
            self._set_link_color(body)

        if len(self._dof_indices[body.name]) > 0 and any(
            not body.attr_array_default_flag[x] and getattr(body, x) is not None
            for x in self._ATTR_DOF_DYNAMICS
        ):
            self._set_dof_dynamics(body)

        if any(
            getattr(body, x) is None
            for x in self._ATTR_LINK_DYNAMICS
            if x not in ("link_linear_damping", "link_angular_damping")
        ):
            dynamics_info = [
                self._p.getDynamicsInfo(self._body_ids[body.name], i)
                for i in range(-1, self._num_links[body.name] - 1)
            ]
            if body.link_lateral_friction is None:
                body.link_lateral_friction = [[x[1] for x in dynamics_info]]
                body.attr_array_default_flag["link_lateral_friction"] = True
            if body.link_spinning_friction is None:
                body.link_spinning_friction = [[x[7] for x in dynamics_info]]
                body.attr_array_default_flag["link_spinning_friction"] = True
            if body.link_rolling_friction is None:
                body.link_rolling_friction = [[x[6] for x in dynamics_info]]
                body.attr_array_default_flag["link_rolling_friction"] = True
            if body.link_restitution is None:
                body.link_restitution = [[x[5] for x in dynamics_info]]
                body.attr_array_default_flag["link_restitution"] = True

        if body.link_color is None:
            visual_data = self._p.getVisualShapeData(self._body_ids[body.name])
            body.link_color = [[x[7] for x in visual_data]]
            body.attr_array_default_flag["link_color"] = True

        if len(self._dof_indices[body.name]) > 0 and any(
            getattr(body, x) is None for x in self._ATTR_DOF_DYNAMICS
        ):
            joint_info = [
                self._p.getJointInfo(self._body_ids[body.name], j)
                for j in self._dof_indices[body.name]
            ]
            if body.dof_lower_limit is None:
                body.dof_lower_limit = [[x[8] for x in joint_info]]
                body.attr_array_default_flag["dof_lower_limit"] = True
            if body.dof_upper_limit is None:
                body.dof_upper_limit = [[x[9] for x in joint_info]]
                body.attr_array_default_flag["dof_upper_limit"] = True

        body.lock_attr_array()

        for attr in (
            ("link_color", "link_collision_filter")
            + self._ATTR_LINK_DYNAMICS
            + self._ATTR_DOF_DYNAMICS
        ):
            if body.attr_array_dirty_flag[attr]:
                body.attr_array_dirty_flag[attr] = False

    def _set_link_collision_filter(self, body):
        """ """
        link_collision_filter = body.get_attr_array("link_collision_filter", 0)
        if (
            not body.attr_array_locked["link_collision_filter"]
            and len(link_collision_filter) != self._num_links[body.name]
        ):
            raise ValueError(
                "Size of 'link_collision_filter' in the link dimension "
                f"({len(link_collision_filter)}) should match the number of links "
                f"({self._num_links[body.name]}): '{body.name}'"
            )
        for i in range(-1, self._num_links[body.name] - 1):
            self._p.setCollisionFilterGroupMask(
                self._body_ids[body.name],
                i,
                link_collision_filter[i + 1],
                link_collision_filter[i + 1],
            )

    def _set_link_dynamics(self, body, dirty_only=False):
        """ """
        for attr in self._ATTR_LINK_DYNAMICS:
            if attr in ("link_linear_damping", "link_angular_damping"):
                continue
            if (
                not body.attr_array_locked[attr]
                and getattr(body, attr) is not None
                and len(body.get_attr_array(attr, 0)) != self._num_links[body.name]
            ):
                raise ValueError(
                    f"Size of '{attr}' in the link dimension ({len(body.get_attr_array(attr, 0))}) "
                    f"should match the number of links ({self._num_links[body.name]}): "
                    f"'{body.name}'"
                )
        kwargs = {}
        if (
            not dirty_only
            and not body.attr_array_default_flag["link_lateral_friction"]
            and body.link_lateral_friction is not None
            or body.attr_array_dirty_flag["link_lateral_friction"]
        ):
            kwargs["lateralFriction"] = body.get_attr_array("link_lateral_friction", 0)
        if (
            not dirty_only
            and not body.attr_array_default_flag["link_spinning_friction"]
            and body.link_spinning_friction is not None
            or body.attr_array_dirty_flag["link_spinning_friction"]
        ):
            kwargs["spinningFriction"] = body.get_attr_array("link_spinning_friction", 0)
        if (
            not dirty_only
            and not body.attr_array_default_flag["link_rolling_friction"]
            and body.link_rolling_friction is not None
            or body.attr_array_dirty_flag["link_rolling_friction"]
        ):
            kwargs["rollingFriction"] = body.get_attr_array("link_rolling_friction", 0)
        if (
            not dirty_only
            and not body.attr_array_default_flag["link_restitution"]
            and body.link_restitution is not None
            or body.attr_array_dirty_flag["link_restitution"]
        ):
            kwargs["restitution"] = body.get_attr_array("link_restitution", 0)
        if len(kwargs) > 0:
            for i in range(-1, self._num_links[body.name] - 1):
                self._p.changeDynamics(
                    self._body_ids[body.name], i, **{k: v[i + 1] for k, v in kwargs.items()}
                )
        # Bullet only sets `linearDamping` and `angularDamping` for link index -1. See:
        #     https://github.com/bulletphysics/bullet3/blob/740d2b978352b16943b24594572586d95d476466/examples/SharedMemory/PhysicsClientC_API.cpp#L3419
        #     https://github.com/bulletphysics/bullet3/blob/740d2b978352b16943b24594572586d95d476466/examples/SharedMemory/PhysicsClientC_API.cpp#L3430
        kwargs = {}
        if (
            not dirty_only
            and not body.attr_array_default_flag["link_linear_damping"]
            and body.link_linear_damping is not None
            or body.attr_array_dirty_flag["link_linear_damping"]
        ):
            kwargs["linearDamping"] = body.get_attr_array("link_linear_damping", 0)
        if (
            not dirty_only
            and not body.attr_array_default_flag["link_angular_damping"]
            and body.link_angular_damping is not None
            or body.attr_array_dirty_flag["link_angular_damping"]
        ):
            kwargs["angularDamping"] = body.get_attr_array("link_angular_damping", 0)
        if len(kwargs) > 0:
            self._p.changeDynamics(
                self._body_ids[body.name], -1, **{k: v for k, v in kwargs.items()}
            )

    def _set_link_color(self, body):
        """ """
        link_color = body.get_attr_array("link_color", 0)
        if (
            not body.attr_array_locked["link_color"]
            and len(link_color) != self._num_links[body.name]
        ):
            raise ValueError(
                f"Size of 'link_color' in the link dimension ({len(link_color)}) should match the "
                f"number of links: '{body.name}' ({self._num_links[body.name]})"
            )
        for i in range(-1, self._num_links[body.name] - 1):
            self._p.changeVisualShape(self._body_ids[body.name], i, rgbaColor=link_color[i + 1])

    def _set_dof_dynamics(self, body, dirty_only=False):
        """ """
        for attr in self._ATTR_DOF_DYNAMICS:
            if (
                not body.attr_array_locked[attr]
                and getattr(body, attr) is not None
                and len(body.get_attr_array(attr, 0)) != len(self._dof_indices[body.name])
            ):
                raise ValueError(
                    f"Size of '{attr}' in the DoF dimension ({len(body.get_attr_array(attr, 0))}) "
                    f"should match the number of DoFs ({len(self._dof_indices[body.name])}): "
                    f"'{body.name}'"
                )
        kwargs = {}
        if (
            not dirty_only
            and not body.attr_array_default_flag["dof_lower_limit"]
            and body.dof_lower_limit is not None
            or body.attr_array_dirty_flag["dof_lower_limit"]
        ):
            kwargs["jointLowerLimit"] = body.get_attr_array("dof_lower_limit", 0)
        if (
            not dirty_only
            and not body.attr_array_default_flag["dof_upper_limit"]
            and body.dof_upper_limit is not None
            or body.attr_array_dirty_flag["dof_upper_limit"]
        ):
            kwargs["jointUpperLimit"] = body.get_attr_array("dof_upper_limit", 0)
        if len(kwargs) > 0:
            for i, j in enumerate(self._dof_indices[body.name]):
                self._p.changeDynamics(
                    self._body_ids[body.name], j, **{k: v[i] for k, v in kwargs.items()}
                )

    def _set_callback_body(self, body):
        """ """
        body.set_callback_collect_dof_state(self._collect_dof_state)
        body.set_callback_collect_link_state(self._collect_link_state)

    def _collect_dof_state(self, body):
        """ """
        if len(self._dof_indices[body.name]) > 0:
            joint_states = self._p.getJointStates(
                self._body_ids[body.name], self._dof_indices[body.name]
            )
            dof_state = [x[0:2] for x in joint_states]
            body.dof_state = [dof_state]

    def _collect_link_state(self, body):
        """ """
        pos, orn = self._p.getBasePositionAndOrientation(self._body_ids[body.name])
        lin, ang = self._p.getBaseVelocity(self._body_ids[body.name])
        link_state = [pos + orn + lin + ang]
        if self._num_links[body.name] > 1:
            link_indices = [*range(0, self._num_links[body.name] - 1)]
            # Need to set computeForwardKinematics=1. See:
            #     https://github.com/bulletphysics/bullet3/issues/2806
            link_states = self._p.getLinkStates(
                self._body_ids[body.name],
                link_indices,
                computeLinkVelocity=1,
                computeForwardKinematics=1,
            )
            link_state += [x[4] + x[5] + x[6] + x[7] for x in link_states]
        body.link_state = [link_state]

    def _load_camera(self, camera):
        """ """
        self._set_projection_matrix(camera)
        self._set_view_matrix(camera)

        self._image_cache[camera.name] = {}
        self._clear_image_cache(camera)

        camera.lock_attr_array()

    def _set_projection_matrix(self, camera):
        """ """
        self._projection_matrix[camera.name] = self._p.computeProjectionMatrixFOV(
            camera.get_attr_array("vertical_fov", 0),
            camera.get_attr_array("width", 0) / camera.get_attr_array("height", 0),
            camera.get_attr_array("near", 0),
            camera.get_attr_array("far", 0),
        )

    def _set_view_matrix(self, camera):
        """ """
        if camera.target is not None and camera.up_vector is not None:
            self._view_matrix[camera.name] = self._p.computeViewMatrix(
                camera.get_attr_array("position", 0),
                camera.get_attr_array("target", 0),
                camera.get_attr_array("up_vector", 0),
            )
        elif camera.orientation is not None:
            R = np.array(
                self._p.getMatrixFromQuaternion(camera.get_attr_array("orientation", 0)),
                dtype=np.float32,
            ).reshape(3, 3)
            view_matrix = np.eye(4, dtype=np.float32)
            view_matrix[:3, :3] = R
            view_matrix[3, :3] = -1 * camera.get_attr_array("position", 0).dot(R)
            self._view_matrix[camera.name] = tuple(view_matrix.flatten())
        else:
            raise ValueError(
                "For Bullet, either ('target', 'up_vector') or 'orientation' is required to be "
                f"set: {camera.name}"
            )

    def _clear_image_cache(self, camera):
        """ """
        self._image_cache[camera.name]["color"] = None
        self._image_cache[camera.name]["depth"] = None
        self._image_cache[camera.name]["segmentation"] = None

    def _set_callback_camera(self, camera):
        """ """
        camera.set_callback_render_color(self._render_color)
        camera.set_callback_render_depth(self._render_depth)
        camera.set_callback_render_segmentation(self._render_segmentation)

    def _render_color(self, camera):
        """ """
        for body in self._scene.bodies:
            if body.attr_array_dirty_flag["link_color"]:
                self._set_link_color(body)
                body.attr_array_dirty_flag["link_color"] = False
                if self._image_cache[camera.name]["color"] is not None:
                    self._image_cache[camera.name]["color"] = None

        self._check_and_update_camera(camera)
        if self._image_cache[camera.name]["color"] is None:
            self._render(camera)
        camera.color = self._image_cache[camera.name]["color"][None]

    def _render_depth(self, camera):
        """ """
        self._check_and_update_camera(camera)
        if self._image_cache[camera.name]["depth"] is None:
            self._render(camera)
        depth = (
            camera.get_attr_array("far", 0)
            * camera.get_attr_array("near", 0)
            / (
                camera.get_attr_array("far", 0)
                - (camera.get_attr_array("far", 0) - camera.get_attr_array("near", 0))
                * self._image_cache[camera.name]["depth"]
            )
        )
        depth[self._image_cache[camera.name]["depth"] == 1.0] = 0.0
        camera.depth = depth[None]

    def _render_segmentation(self, camera):
        """ """
        for body in self._scene.bodies:
            for attr in ("link_segmentation_id",):
                if getattr(body, attr) is not None:
                    raise ValueError(f"'{attr}' is not supported in Bullet: '{body.name}'")

        self._check_and_update_camera(camera)
        if self._image_cache[camera.name]["segmentation"] is None:
            self._render(camera)
        camera.segmentation = self._image_cache[camera.name]["segmentation"][None]

    def _check_and_update_camera(self, camera):
        """ """
        if any(camera.attr_array_dirty_flag[x] for x in self._ATTR_PROJECTION_MATRIX):
            self._set_projection_matrix(camera)
            for attr in self._ATTR_PROJECTION_MATRIX:
                if camera.attr_array_dirty_flag[attr]:
                    camera.attr_array_dirty_flag[attr] = False
            self._clear_image_cache(camera)
        if any(camera.attr_array_dirty_flag[x] for x in self._ATTR_VIEW_MATRIX):
            self._set_view_matrix(camera)
            for attr in self._ATTR_VIEW_MATRIX:
                if camera.attr_array_dirty_flag[attr]:
                    camera.attr_array_dirty_flag[attr] = False
            self._clear_image_cache(camera)

    def _render(self, camera):
        """ """
        (
            _,
            _,
            self._image_cache[camera.name]["color"],
            self._image_cache[camera.name]["depth"],
            self._image_cache[camera.name]["segmentation"],
        ) = self._p.getCameraImage(
            camera.get_attr_array("width", 0),
            camera.get_attr_array("height", 0),
            viewMatrix=self._view_matrix[camera.name],
            projectionMatrix=self._projection_matrix[camera.name],
            renderer=pybullet.ER_BULLET_HARDWARE_OPENGL,
        )

    def _set_callback_scene(self):
        """ """
        self._scene.set_callback_add_camera(self._add_camera)
        self._scene.set_callback_remove_camera(self._remove_camera)

    def _add_camera(self, camera):
        """ """
        self._load_camera(camera)
        self._set_callback_camera(camera)

    def _remove_camera(self, camera):
        """ """
        del self._projection_matrix[camera.name]
        del self._view_matrix[camera.name]
        del self._image_cache[camera.name]

    def _set_viewer_camera_pose(self, position, target):
        """ """
        disp = [x - y for x, y in zip(position, target)]
        dist = np.linalg.norm(disp)
        yaw = np.arctan2(disp[0], -disp[1])
        yaw = np.rad2deg(yaw)
        pitch = np.arctan2(-disp[2], np.linalg.norm((disp[0], disp[1])))
        pitch = np.rad2deg(pitch)

        self._p.resetDebugVisualizerCamera(dist, yaw, pitch, target)

    def _clear_state(self):
        """ """
        for body in self._scene.bodies:
            body.dof_state = None
            body.link_state = None

    def _clear_image(self):
        """ """
        for camera in self._scene.cameras:
            camera.color = None
            camera.depth = None
            camera.segmentation = None

            self._clear_image_cache(camera)

    def step(self):
        """ """
        if self._cfg.RENDER:
            # Simulate real-time rendering with sleep if computation takes less than real time.
            time_spent = time.time() - self._last_frame_time
            time_sleep = self._cfg.TIME_STEP - time_spent
            if time_sleep > 0:
                time.sleep(time_sleep)
            self._last_frame_time = time.time()

        for body in reversed(self._scene_cache.bodies):
            if body.name not in [x.name for x in self._scene.bodies]:
                # Remove body.
                with self._disable_cov_rendering():
                    self._p.removeBody(self._body_ids[body.name])
                    del self._body_ids[body.name]
                    del self._dof_indices[body.name]
                    del self._num_links[body.name]
                    self._scene_cache.remove_body(body)
        for body in self._scene.bodies:
            if body.name not in [x.name for x in self._scene_cache.bodies]:
                # Add body.
                with self._disable_cov_rendering():
                    self._load_body(body)
                    self._cache_body_and_set_control_and_props(body)
                    self._set_callback_body(body)

        for body in self._scene.bodies:
            for attr in (
                "link_segmentation_id",
                "dof_has_limits",
                "dof_armature",
            ):
                if getattr(body, attr) is not None:
                    raise ValueError(f"'{attr}' is not supported in Bullet: '{body.name}'")

            if body.env_ids_load is not None and len(body.env_ids_load) == 0:
                for attr in ("env_ids_reset_base_state", "env_ids_reset_dof_state"):
                    if getattr(body, attr) is not None:
                        raise ValueError(
                            f"For Bullet, '{attr}' should be None if 'env_ids_load' is set to []: "
                            f"'{body.name}'"
                        )
                continue

            if body.env_ids_reset_base_state is not None:
                if not np.array_equal(body.env_ids_reset_base_state.cpu(), [0]):
                    raise ValueError(
                        "For Bullet, 'env_ids_reset_base_state' must be either None or [0]: "
                        f"'{body.name}'"
                    )
                if body.initial_base_position is None and body.initial_base_velocity is None:
                    raise ValueError(
                        "'initial_base_position' and 'initial_base_velocity' cannot be both None "
                        f"when 'env_ids_reset_base_state' is used: {body.name}"
                    )
                if body.initial_base_position is not None:
                    self._reset_base_position(body)
                if body.initial_base_velocity is not None:
                    self._reset_base_velocity(body)
                body.env_ids_reset_base_state = None

            if body.attr_array_dirty_flag["scale"]:
                raise ValueError(
                    f"For Bullet, 'scale' can only be changed before each reset: '{body.name}'"
                )
            if body.attr_array_dirty_flag["link_collision_filter"]:
                self._set_link_collision_filter(body)
                body.attr_array_dirty_flag["link_collision_filter"] = False
            if any(body.attr_array_dirty_flag[x] for x in self._ATTR_LINK_DYNAMICS):
                self._set_link_dynamics(body, dirty_only=True)
                for attr in self._ATTR_LINK_DYNAMICS:
                    if body.attr_array_dirty_flag[attr]:
                        body.attr_array_dirty_flag[attr] = False
            if body.attr_array_dirty_flag["link_color"]:
                self._set_link_color(body)
                body.attr_array_dirty_flag["link_color"] = False

            if len(self._dof_indices[body.name]) == 0:
                for attr in (
                    "dof_lower_limit",
                    "dof_upper_limit",
                    "dof_control_mode",
                    "dof_max_force",
                    "dof_max_velocity",
                    "dof_position_gain",
                    "dof_velocity_gain",
                    "dof_target_position",
                    "dof_target_velocity",
                    "dof_force",
                    "env_ids_reset_dof_state",
                ):
                    if getattr(body, attr) is not None:
                        raise ValueError(
                            f"'{attr}' must be None for body with 0 DoF: '{body.name}'"
                        )
                continue

            if body.env_ids_reset_dof_state is not None:
                if not np.array_equal(body.env_ids_reset_dof_state.cpu(), [0]):
                    raise ValueError(
                        "For Bullet, 'env_ids_reset_dof_state' must be either None or [0]: "
                        f"'{body.name}'"
                    )
                self._reset_dof_state(body)
                body.env_ids_reset_dof_state = None

            if any(body.attr_array_dirty_flag[x] for x in self._ATTR_DOF_DYNAMICS):
                self._set_dof_dynamics(body, dirty_only=True)
                for attr in self._ATTR_DOF_DYNAMICS:
                    if body.attr_array_dirty_flag[attr]:
                        body.attr_array_dirty_flag[attr] = False

            if body.attr_array_dirty_flag["dof_control_mode"]:
                raise ValueError(
                    "For Bullet, 'dof_control_mode' cannot be changed after each reset: "
                    f"'{body.name}'"
                )
            # The redundant if-else block below is an artifact due to `setJointMotorControlArray()`
            # not supporting `maxVelocity`. `setJointMotorControlArray()` is still preferred when
            # `maxVelocity` is not needed due to better speed performance.
            if body.dof_max_velocity is None:
                kwargs = {}
                if body.dof_target_position is not None:
                    kwargs["targetPositions"] = body.get_attr_tensor("dof_target_position", 0)
                if body.dof_target_velocity is not None:
                    kwargs["targetVelocities"] = body.get_attr_tensor("dof_target_velocity", 0)
                if body.dof_position_gain is not None:
                    kwargs["positionGains"] = body.get_attr_array("dof_position_gain", 0)
                if body.dof_velocity_gain is not None:
                    kwargs["velocityGains"] = body.get_attr_array("dof_velocity_gain", 0)
                if body.dof_control_mode.ndim == 0:
                    if body.dof_max_force is not None:
                        if body.dof_control_mode not in (
                            DoFControlMode.POSITION_CONTROL,
                            DoFControlMode.VELOCITY_CONTROL,
                        ):
                            raise ValueError(
                                "For Bullet, 'dof_max_force' can only be set in POSITION_CONTROL "
                                f"and VELOCITY_CONTROL modes: '{body.name}'"
                            )
                        kwargs["forces"] = body.get_attr_array("dof_max_force", 0)
                    if body.dof_force is not None:
                        if body.dof_control_mode != DoFControlMode.TORQUE_CONTROL:
                            raise ValueError(
                                "For Bullet, 'dof_force' can only be set in the TORQUE_CONTROL "
                                f"mode: '{body.name}'"
                            )
                        kwargs["forces"] = body.get_attr_tensor("dof_force", 0)
                    self._p.setJointMotorControlArray(
                        self._body_ids[body.name],
                        self._dof_indices[body.name],
                        self._DOF_CONTROL_MODE_MAP[body.dof_control_mode.item()],
                        **kwargs,
                    )
                if body.dof_control_mode.ndim == 1:
                    if body.dof_max_force is not None:
                        if (
                            DoFControlMode.POSITION_CONTROL not in body.dof_control_mode
                            and DoFControlMode.VELOCITY_CONTROL not in body.dof_control_mode
                        ):
                            raise ValueError(
                                "For Bullet, 'dof_max_force' can only be set in POSITION_CONTROL "
                                f"and VELOCITY_CONTROL modes: '{body.name}'"
                            )
                        kwargs["forces"] = body.get_attr_array("dof_max_force", 0)
                    if DoFControlMode.POSITION_CONTROL in body.dof_control_mode:
                        self._p.setJointMotorControlArray(
                            self._body_ids[body.name],
                            self._dof_indices[body.name][
                                body.dof_control_mode == DoFControlMode.POSITION_CONTROL
                            ],
                            self._DOF_CONTROL_MODE_MAP[DoFControlMode.POSITION_CONTROL],
                            **{
                                k: v[body.dof_control_mode == DoFControlMode.POSITION_CONTROL]
                                for k, v in kwargs.items()
                            },
                        )
                    if DoFControlMode.VELOCITY_CONTROL in body.dof_control_mode:
                        self._p.setJointMotorControlArray(
                            self._body_ids[body.name],
                            self._dof_indices[body.name][
                                body.dof_control_mode == DoFControlMode.VELOCITY_CONTROL
                            ],
                            self._DOF_CONTROL_MODE_MAP[DoFControlMode.VELOCITY_CONTROL],
                            **{
                                k: v[body.dof_control_mode == DoFControlMode.VELOCITY_CONTROL]
                                for k, v in kwargs.items()
                            },
                        )
                    if "forces" in kwargs:
                        del kwargs["forces"]
                    if body.dof_force is not None:
                        if DoFControlMode.TORQUE_CONTROL not in body.dof_control_mode:
                            raise ValueError(
                                "For Bullet, 'dof_force' can only be set in the TORQUE_CONTROL "
                                f"mode: '{body.name}'"
                            )
                        kwargs["forces"] = body.get_attr_tensor("dof_force", 0)
                    if DoFControlMode.TORQUE_CONTROL in body.dof_control_mode:
                        self._p.setJointMotorControlArray(
                            self._body_ids[body.name],
                            self._dof_indices[body.name][
                                body.dof_control_mode == DoFControlMode.TORQUE_CONTROL
                            ],
                            self._DOF_CONTROL_MODE_MAP[DoFControlMode.TORQUE_CONTROL],
                            **{
                                k: v[body.dof_control_mode == DoFControlMode.TORQUE_CONTROL]
                                for k, v in kwargs.items()
                            },
                        )
            else:
                kwargs = {}
                if body.dof_target_position is not None:
                    kwargs["targetPosition"] = body.get_attr_tensor("dof_target_position", 0)
                if body.dof_target_velocity is not None:
                    kwargs["targetVelocity"] = body.get_attr_tensor("dof_target_velocity", 0)
                if body.dof_max_velocity is not None:
                    # For Bullet, 'dof_max_velocity' has no effect when not in the POSITION_CONROL
                    # mode.
                    kwargs["maxVelocity"] = body.get_attr_array("dof_max_velocity", 0)
                if body.dof_position_gain is not None:
                    kwargs["positionGain"] = body.get_attr_array("dof_position_gain", 0)
                if body.dof_velocity_gain is not None:
                    kwargs["velocityGain"] = body.get_attr_array("dof_velocity_gain", 0)
                if body.dof_control_mode.ndim == 0:
                    if body.dof_max_force is not None:
                        if body.dof_control_mode not in (
                            DoFControlMode.POSITION_CONTROL,
                            DoFControlMode.VELOCITY_CONTROL,
                        ):
                            raise ValueError(
                                "For Bullet, 'dof_max_force' can only be set in POSITION_CONTROL "
                                f"and VELOCITY_CONTROL modes: '{body.name}'"
                            )
                        kwargs["force"] = body.get_attr_array("dof_max_force", 0)
                    if body.dof_force is not None:
                        if body.dof_control_mode != DoFControlMode.TORQUE_CONTROL:
                            raise ValueError(
                                "For Bullet, 'dof_force' can only be set in the TORQUE_CONTROL "
                                f"mode: '{body.name}'"
                            )
                        kwargs["force"] = body.get_attr_tensor("dof_force", 0)
                    for i, j in enumerate(self._dof_indices[body.name]):
                        self._p.setJointMotorControl2(
                            self._body_ids[body.name],
                            j,
                            self._DOF_CONTROL_MODE_MAP[body.dof_control_mode.item()],
                            **{k: v[i] for k, v in kwargs.items()},
                        )
                if body.dof_control_mode.ndim == 1:
                    if (
                        body.dof_max_force is not None
                        and DoFControlMode.POSITION_CONTROL not in body.dof_control_mode
                        and DoFControlMode.VELOCITY_CONTROL not in body.dof_control_mode
                    ):
                        raise ValueError(
                            "For Bullet, 'dof_max_force' can only be set in POSITION_CONTROL and "
                            f"VELOCITY_CONTROL modes: '{body.name}'"
                        )
                    if (
                        body.dof_force is not None
                        and DoFControlMode.TORQUE_CONTROL not in body.dof_control_mode
                    ):
                        raise ValueError(
                            "For Bullet, 'dof_force' can only be set in the TORQUE_CONTROL mode: "
                            f"'{body.name}'"
                        )
                    for i, j in enumerate(self._dof_indices[body.name]):
                        if body.dof_control_mode[i] in (
                            DoFControlMode.POSITION_CONTROL,
                            DoFControlMode.VELOCITY_CONTROL,
                        ):
                            if "force" in kwargs:
                                del kwargs["force"]
                            if body.dof_max_force is not None:
                                kwargs["force"] = body.get_attr_array("dof_max_force", 0)
                            if body.dof_force is not None and not np.isnan(
                                body.get_attr_tensor("dof_force", 0)[i]
                            ):
                                raise ValueError(
                                    "For Bullet, 'dof_force' is required to be np.nan for DoF "
                                    f"({i}) in POSITION_CONTROL and VELOCITY modes: {body.name}"
                                )
                            self._p.setJointMotorControl2(
                                self._body_ids[body.name],
                                j,
                                self._DOF_CONTROL_MODE_MAP[body.dof_control_mode[i].item()],
                                **{k: v[i] for k, v in kwargs.items()},
                            )
                        if body.dof_control_mode[i] == DoFControlMode.TORQUE_CONTROL:
                            if "force" in kwargs:
                                del kwargs["force"]
                            if body.dof_force is not None:
                                kwargs["force"] = body.get_attr_tensor("dof_force", 0)
                            if body.dof_max_force is not None and not np.isnan(
                                body.get_attr_array("dof_max_force", 0)[i]
                            ):
                                raise ValueError(
                                    "For Bullet, 'dof_max_force' is required to be np.nan for DoF "
                                    f"({i}) in the TORQUE_CONTROL mode: {body.name}"
                                )
                            self._p.setJointMotorControl2(
                                self._body_ids[body.name],
                                j,
                                self._DOF_CONTROL_MODE_MAP[body.dof_control_mode[i].item()],
                                **{k: v[i] for k, v in kwargs.items()},
                            )

        self._p.stepSimulation()

        self._clear_state()
        self._clear_image()
        self._contact = None

    @property
    def contact(self):
        """ """
        if self._contact is None:
            self._contact = self._collect_contact()
        return self._contact

    def _collect_contact(self):
        """ """
        pts = self._p.getContactPoints()
        if len(pts) == 0:
            contact_array = create_contact_array(0)
        else:
            kwargs = {}
            kwargs["body_id_a"] = [x[1] if x[1] != self._body_id_ground_plane else -1 for x in pts]
            kwargs["body_id_b"] = [x[2] if x[2] != self._body_id_ground_plane else -1 for x in pts]
            kwargs["link_id_a"] = [x[3] + 1 for x in pts]
            kwargs["link_id_b"] = [x[4] + 1 for x in pts]
            kwargs["position_a_world"] = [x[5] for x in pts]
            kwargs["position_b_world"] = [x[6] for x in pts]
            kwargs["position_a_link"] = np.nan
            kwargs["position_b_link"] = np.nan
            kwargs["normal"] = [x[7] for x in pts]
            kwargs["force"] = [x[9] for x in pts]
            contact_array = create_contact_array(len(pts), **kwargs)
        return [contact_array]

    def close(self):
        """ """
        if self._connected:
            for name in self._plugins:
                self._p.unloadPlugin(self._plugins[name])
            self._p.disconnect()
            self._connected = False



# Copyright (c) 2021, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# Licensed under the MIT License [see LICENSE for details].

import importlib


def load(entry_point):
    """ """
    mod_name, attr_name = entry_point.split(":")
    mod = importlib.import_module(mod_name)
    fn = getattr(mod, attr_name)
    return fn


class SimulatorSpec:
    """ """

    def __init__(self, name, entry_point, kwargs=None):
        """ """
        self._name = name
        self._entry_point = entry_point
        self._kwargs = {} if kwargs is None else kwargs

    def make(self, **kwargs):
        """ """
        _kwargs = self._kwargs.copy()
        _kwargs.update(kwargs)

        fn = load(self._entry_point)
        simulator = fn(**_kwargs)

        return simulator


class SimulatorRegistry:
    """ """

    def __init__(self):
        """ """
        self._specs = {}

    def register(self, name, **kwargs):
        """ """
        if name in self._specs:
            raise Exception(f"Cannot re-register name: '{name}'")
        self._specs[name] = SimulatorSpec(name, **kwargs)

    def make(self, name, **kwargs):
        """ """
        if name not in self._specs:
            raise KeyError(f"No registered simulator with name: '{name}'")
        spec = self._specs[name]

        simulator = spec.make(**kwargs)

        return simulator


registry = SimulatorRegistry()


def register(name, **kwargs):
    """ """
    registry.register(name, **kwargs)


def make(name, **kwargs):
    """ """
    return registry.make(name, **kwargs)


register(
    name="bullet",
    entry_point="easysim.simulators.bullet:Bullet",
)

register(
    name="isaac_gym",
    entry_point="easysim.simulators.isaac_gym:IsaacGym",
)
